//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Histo/Histogram1D.cpp
//! @brief     Implements class Histogram1D.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Histo/Histogram1D.h"
#include "Base/Axis/FixedBinAxis.h"
#include "Base/Axis/VariableBinAxis.h"
#include "Base/Py/PyUtils.h"
#include "Device/Data/Powerfield.h"
#include <memory>

Histogram1D::Histogram1D(int nbinsx, double xlow, double xup)
    : IHistogram(FixedBinAxis("x-axis", nbinsx, xlow, xup))
{
}

Histogram1D::Histogram1D(int nbinsx, const std::vector<double>& xbins)
    : IHistogram(VariableBinAxis("x-axis", nbinsx, xbins))
{
}

Histogram1D::Histogram1D(const IAxis& axis)
    : IHistogram(axis)
{
}

Histogram1D::Histogram1D(const Powerfield<double>& data)
{
    init_from_data(data);
}

Histogram1D* Histogram1D::clone() const
{
    return new Histogram1D(*this);
}

int Histogram1D::fill(double x, double weight)
{
    const IAxis& axis = xAxis();
    if (!axis.contains(x))
        return -1;
    size_t index = axis.findClosestIndex(x);
    (*m_data)[index].add(weight);
    return (int)index;
}

std::vector<double> Histogram1D::binCenters() const
{
    return xAxis().binCenters();
}

std::vector<double> Histogram1D::binValues() const
{
    return IHistogram::getDataVector(IHistogram::DataType::INTEGRAL);
}

std::vector<double> Histogram1D::binErrors() const
{
    return IHistogram::getDataVector(IHistogram::DataType::STANDARD_ERROR);
}

#ifdef BORNAGAIN_PYTHON

PyObject* Histogram1D::binCentersNumpy() const
{
    return BaseUtils::Python::createNumpyArray(binCenters());
}

PyObject* Histogram1D::binValuesNumpy() const
{
    return BaseUtils::Python::createNumpyArray(binValues());
}

PyObject* Histogram1D::binErrorsNumpy() const
{
    return BaseUtils::Python::createNumpyArray(binErrors());
}

#endif // BORNAGAIN_PYTHON

Histogram1D* Histogram1D::crop(double xmin, double xmax)
{
    const std::unique_ptr<IAxis> xaxis{xAxis().clone()};
    xaxis->clip(xmin, xmax);
    auto* result = new Histogram1D(*xaxis);
    Powerfield<CumulativeValue>::const_iterator it_origin = m_data->begin();
    Powerfield<CumulativeValue>::iterator it_result = result->m_data->begin();
    while (it_origin != m_data->end()) {
        const double x = m_data->projectedCoord(it_origin.getIndex(), 0);
        if (result->xAxis().contains(x)) {
            *it_result = *it_origin;
            ++it_result;
        }
        ++it_origin;
    }
    return result;
}
