#include "Device/Detector/SphericalDetector.h"
#include "Base/Const/Units.h"
#include "Device/Beam/Beam.h"
#include "Device/Mask/DetectorMask.h"
#include "Device/Mask/Polygon.h"
#include "Device/Mask/Rectangle.h"
#include "Device/Resolution/ConvolutionDetectorResolution.h"
#include "Device/Resolution/ResolutionFunction2DGaussian.h"
#include "Tests/GTestWrapper/google_test.h"
#include <memory>

class SphericalDetectorTest : public ::testing::Test {
};

// Construction of the detector with axes.
TEST_F(SphericalDetectorTest, constructionWithAxes)
{
    SphericalDetector detector;
    FixedBinAxis axis0("axis0", 10, 0.0, 10.0);
    FixedBinAxis axis1("axis1", 20, 0.0, 20.0);
    detector.addDetAxis(axis0);
    detector.addDetAxis(axis1);

    // checking dimension and axes
    EXPECT_EQ(2u, detector.rank());
    EXPECT_EQ(axis0.min(), detector.axis(0).min());
    EXPECT_EQ(axis0.max(), detector.axis(0).max());
    EXPECT_EQ(axis1.min(), detector.axis(1).min());
    EXPECT_EQ(axis1.max(), detector.axis(1).max());
}

// Construction of the detector via classical constructor.
TEST_F(SphericalDetectorTest, constructionWithParameters)
{
    SphericalDetector detector(10, -1.0, 1.0, 20, 0.0, 2.0);
    EXPECT_EQ(10u, detector.axis(0).size());
    EXPECT_EQ(-1.0, detector.axis(0).min());
    EXPECT_EQ(1.0, detector.axis(0).max());
    EXPECT_EQ(20u, detector.axis(1).size());
    EXPECT_EQ(0.0, detector.axis(1).min());
    EXPECT_EQ(2.0, detector.axis(1).max());
}

// Creation of the detector map with axes in given units
TEST_F(SphericalDetectorTest, createDetectorMap)
{
    SphericalDetector detector(10, -1.0 * Units::deg, 1.0 * Units::deg, 20, 0.0 * Units::deg,
                               2.0 * Units::deg);

    // creating map in default units, which are radians and checking axes
    auto data = detector.createDetectorMap();
    EXPECT_EQ(data->axis(0).size(), 10u);
    EXPECT_EQ(data->axis(0).min(), -1.0 * Units::deg);
    EXPECT_EQ(data->axis(0).max(), 1.0 * Units::deg);
    EXPECT_EQ(data->axis(1).size(), 20u);
    EXPECT_EQ(data->axis(1).min(), 0.0 * Units::deg);
    EXPECT_EQ(data->axis(1).max(), 2.0 * Units::deg);
}

// Testing region of interest.
TEST_F(SphericalDetectorTest, regionOfInterest)
{
    SphericalDetector detector;
    detector.addDetAxis(FixedBinAxis("axis0", 8, -3.0, 5.0));
    detector.addDetAxis(FixedBinAxis("axis1", 4, 0.0, 4.0));

    // creating region of interest
    double xlow(-2.0), ylow(1.0), xup(4.0), yup(3.0);
    detector.setRegionOfInterest(xlow, ylow, xup, yup);
    EXPECT_TRUE(detector.hasExplicitRegionOfInterest());
    EXPECT_EQ(detector.regionOfInterestBounds(0).first, xlow);
    EXPECT_EQ(detector.regionOfInterestBounds(0).second, xup);
    EXPECT_EQ(detector.regionOfInterestBounds(1).first, ylow);
    EXPECT_EQ(detector.regionOfInterestBounds(1).second, yup);

    // replacing region of interest with a new one
    double xlow2(-2.1), ylow2(1.1), xup2(4.1), yup2(3.1);
    detector.setRegionOfInterest(xlow2, ylow2, xup2, yup2);
    EXPECT_EQ(detector.regionOfInterestBounds(0).first, xlow2);
    EXPECT_EQ(detector.regionOfInterestBounds(0).second, xup2);
    EXPECT_EQ(detector.regionOfInterestBounds(1).first, ylow2);
    EXPECT_EQ(detector.regionOfInterestBounds(1).second, yup2);

    // removing region of interest
    detector.resetRegionOfInterest();
    EXPECT_FALSE(detector.hasExplicitRegionOfInterest());
}

// Create detector map in the presence of region of interest.
TEST_F(SphericalDetectorTest, regionOfInterestAndDetectorMap)
{
    SphericalDetector detector(6, -1.0 * Units::deg, 5.0 * Units::deg, 4, 0.0 * Units::deg,
                               4.0 * Units::deg);

    detector.setRegionOfInterest(0.1 * Units::deg, 1.1 * Units::deg, 3.0 * Units::deg,
                                 2.9 * Units::deg);
    // Creating map in default units, which are radians and checking that axes are clipped
    // to region of interest.
    auto data = detector.createDetectorMap();
    EXPECT_EQ(data->axis(0).size(), 4u);
    EXPECT_EQ(data->axis(0).min(), 0.0 * Units::deg);
    EXPECT_EQ(data->axis(0).max(), 4.0 * Units::deg);
    EXPECT_EQ(data->axis(1).size(), 2u);
    EXPECT_EQ(data->axis(1).min(), 1.0 * Units::deg);
    EXPECT_EQ(data->axis(1).max(), 3.0 * Units::deg);
}

TEST_F(SphericalDetectorTest, MaskOfDetector)
{
    SphericalDetector detector;
    detector.addDetAxis(FixedBinAxis("x-axis", 12, -4.0, 8.0));
    detector.addDetAxis(FixedBinAxis("y-axis", 6, -2.0, 4.0));

    std::vector<double> x = {4.0, -4.0, -4.0, 4.0, 4.0};
    std::vector<double> y = {2.0, 2.0, -2.0, -2.0, 2.0};

    Polygon polygon(x, y);
    detector.addMask(polygon, true);

    const Powerfield<bool>* mask = detector.detectorMask()->getMaskData();
    for (size_t index = 0; index < mask->allocatedSize(); ++index) {
        double x = mask->projectedCoord(index, 0);
        double y = mask->projectedCoord(index, 1);
        if (x >= -4.0 && x <= 4.0 && y >= -2.0 && y <= 2.0)
            EXPECT_TRUE(detector.detectorMask()->isMasked(index));
        else
            EXPECT_FALSE(detector.detectorMask()->isMasked(index));
    }

    SphericalDetector detector2(detector);
    mask = detector2.detectorMask()->getMaskData();
    for (size_t index = 0; index < mask->allocatedSize(); ++index) {
        double x = mask->projectedCoord(index, 0);
        double y = mask->projectedCoord(index, 1);
        if (x >= -4.0 && x <= 4.0 && y >= -2.0 && y <= 2.0)
            EXPECT_TRUE(detector2.detectorMask()->isMasked(index));
        else
            EXPECT_FALSE(detector2.detectorMask()->isMasked(index));
    }

    mask = detector.detectorMask()->getMaskData();
    for (size_t index = 0; index < mask->allocatedSize(); ++index) {
        double x = mask->projectedCoord(index, 0);
        double y = mask->projectedCoord(index, 1);
        if (x >= -4.0 && x <= 4.0 && y >= -2.0 && y <= 2.0)
            EXPECT_TRUE(detector.detectorMask()->isMasked(index));
        else
            EXPECT_FALSE(detector.detectorMask()->isMasked(index));
    }
}

// Checking clone in the presence of ROI and masks.
TEST_F(SphericalDetectorTest, Clone)
{
    SphericalDetector detector(6, -1.0 * Units::deg, 5.0 * Units::deg, 4, 0.0 * Units::deg,
                               4.0 * Units::deg);
    detector.setRegionOfInterest(0.1 * Units::deg, 1.1 * Units::deg, 3.0 * Units::deg,
                                 2.9 * Units::deg);
    detector.addMask(
        Rectangle(-0.9 * Units::deg, 0.1 * Units::deg, 0.9 * Units::deg, 1.9 * Units::deg), true);
    detector.addMask(
        Rectangle(3.1 * Units::deg, 2.1 * Units::deg, 4.9 * Units::deg, 3.9 * Units::deg), true);
    detector.setDetectorResolution(
        ConvolutionDetectorResolution(ResolutionFunction2DGaussian(1, 1)));

    std::unique_ptr<SphericalDetector> clone(detector.clone());

    auto data = clone->createDetectorMap();
    EXPECT_EQ(data->axis(0).size(), 4u);
    EXPECT_EQ(data->axis(0).min(), 0.0 * Units::deg);
    EXPECT_EQ(data->axis(0).max(), 4.0 * Units::deg);
    EXPECT_EQ(data->axis(1).size(), 2u);
    EXPECT_EQ(data->axis(1).min(), 1.0 * Units::deg);
    EXPECT_EQ(data->axis(1).max(), 3.0 * Units::deg);

    EXPECT_EQ(clone->detectorMask()->numberOfMaskedChannels(), 8);

    // checking iteration over the map of cloned detector
    std::vector<size_t> expectedDetectorIndexes = {6, 9, 10, 13, 14, 17};
    std::vector<size_t> detectorIndexes;
    for (auto it = clone->beginNonMaskedPoints(); it != clone->endNonMaskedPoints(); ++it)
        detectorIndexes.push_back(it.detectorIndex());
    EXPECT_EQ(detectorIndexes, expectedDetectorIndexes);
}
